//-----------------------------------------------------------------------------
// app.cpp
//
// Demonstrates RS Line Pattern, RS Edge Antialias, RS Last Pixel.  This
// file includes the boiler-plate and GUI handling code.
//-----------------------------------------------------------------------------
#include <iomanip>
#include <sstream>
#include <vector>

#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <basetsd.h>
#include <math.h>
#include <D3DX8.h>
#include <DXErr8.h>
#include <tchar.h>
#include "D3DApp.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "resource.h"
#include "rt_Rasterize.h"

#include <atlbase.h>

#include "rt/app.h"
#include "rt/hr.h"
#include "rt/tstring.h"
#include "rt/vertexbuf.h"

#include "colorsel.h"
#include "vertices.h"

const UINT NUM_CIRCLE_VERTICES = 200;

//-----------------------------------------------------------------------------
// Global access to the app (needed for the global WndProc())
//-----------------------------------------------------------------------------
CMyD3DApplication* g_pApp  = NULL;
HINSTANCE          g_hInst = NULL;



//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    CMyD3DApplication d3dApp;

    g_pApp  = &d3dApp;
    g_hInst = hInst;

    if( FAILED( d3dApp.Create( hInst ) ) )
        return 0;

    return d3dApp.Run();
}




//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
    : m_show_text(true),
    m_animate(false),
    m_grid_fg(D3DCOLOR_ARGB(100, 255, 255, 0)),
    m_circle_fg(D3DCOLOR_ARGB(100, 0, 255, 255)),
    m_background(D3DCOLOR_XRGB(0, 0, 0)),
    m_line_pattern(0),
    m_last_pixel(true),
    m_antialias_edge(false),
    m_can_antialias_edge(false),
    m_can_pattern_line(false),
    m_show_grid(false),
    m_show_circle(false),
    m_grid(),
    m_num_grid_segments(0),
    m_circle(),
    m_pFont(new CD3DFont(_T("Arial"), 12, D3DFONT_BOLD)),
    m_bLoadingApp(TRUE),
    m_fWorldRotX(0.f),
    m_fWorldRotY(0.f)
{
    m_dwCreationWidth           = 500;
    m_dwCreationHeight          = 375;
    m_strWindowTitle            = TEXT("rt_Rasterize");
    m_bUseDepthBuffer           = FALSE;

    ZeroMemory( &m_UserInput, sizeof(m_UserInput) );

    // Read settings from registry
    ReadSettings();
}




//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    // Drawing loading status message until app finishes loading
    SendMessage( m_hWnd, WM_PAINT, 0, 0 );

    m_bLoadingApp = FALSE;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: ReadSettings()
// Desc: Read the app settings from the registry
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::ReadSettings()
{
    HKEY hkey;
    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, DXAPP_KEY, 
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        // Read the stored window width/height.  This is just an example,
        // of how to use DXUtil_Read*() functions.
        DXUtil_ReadIntRegKey( hkey, TEXT("Width"), &m_dwCreationWidth, m_dwCreationWidth );
        DXUtil_ReadIntRegKey( hkey, TEXT("Height"), &m_dwCreationHeight, m_dwCreationHeight );

        RegCloseKey( hkey );
    }
}




//-----------------------------------------------------------------------------
// Name: WriteSettings()
// Desc: Write the app settings to the registry
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::WriteSettings()
{
    HKEY hkey;
    DWORD dwType = REG_DWORD;
    DWORD dwLength = sizeof(DWORD);

    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, DXAPP_KEY, 
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        // Write the window width/height.  This is just an example,
        // of how to use DXUtil_Write*() functions.
        DXUtil_WriteIntRegKey( hkey, TEXT("Width"), m_rcWindowClient.right );
        DXUtil_WriteIntRegKey( hkey, TEXT("Height"), m_rcWindowClient.bottom );

        RegCloseKey( hkey );
    }
}





//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the display device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
    bool caps_ok = true;

    return caps_ok ? S_OK : E_FAIL;
}




//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
    // Init the font
    m_pFont->InitDeviceObjects( m_pd3dDevice );

    return S_OK;
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::RestoreDeviceObjects()
//
// Adjust GUI state based on possibly new device caps, create grid VB and
// screen-space grid vertices based on back buffer dimensions, create
// circle VB and initialize matrices.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    // adjust menu state based on device capabilities
    m_can_antialias_edge = (m_d3dCaps.RasterCaps & D3DPRASTERCAPS_ANTIALIASEDGES) != 0;
    m_can_pattern_line = (m_d3dCaps.RasterCaps & D3DPRASTERCAPS_PAT) != 0;
    {
        HMENU menu = ::GetMenu(m_hWnd);
        rt::enable_menu(menu, IDM_OPTION_EDGE_ANTIALIAS, m_can_antialias_edge);
        rt::enable_menu(menu, IDM_OPTION_LINE_PATTERN, m_can_pattern_line);
        if (!m_can_antialias_edge)
        {
            m_antialias_edge = false;
            rt::check_menu(menu, IDM_OPTION_EDGE_ANTIALIAS, false);
        }
    }
    if (!m_can_pattern_line)
    {
        m_line_pattern = 0;
    }

    // create grid vertices
    const UINT num_rows = (m_d3dsdBackBuffer.Height+15)/16;
    const UINT num_cols = (m_d3dsdBackBuffer.Width+15)/16;
    m_num_grid_segments = num_rows*(1 + num_cols);
    const UINT num_grid_vertices = m_num_grid_segments*2;
    THR(m_pd3dDevice->CreateVertexBuffer(sizeof(s_grid_vertex)*num_grid_vertices,
        D3DUSAGE_WRITEONLY, s_grid_vertex::FVF, D3DPOOL_MANAGED, &m_grid));
    {
        rt::vertex_lock<s_grid_vertex> lock(m_grid);
        s_grid_vertex *vtx = lock.data();
        UINT i;
        for (i = 0; i < num_rows; i++)
        {
            *vtx = s_grid_vertex(0, i*16.f, m_grid_fg);
            vtx++;
            *vtx = s_grid_vertex(float(m_d3dsdBackBuffer.Width), i*16.f, m_grid_fg);
            vtx++;
        }
        for (i = 0; i < num_rows; i++)
        {
            for (UINT j = 0; j < num_cols; j++)
            {
                *vtx = s_grid_vertex(j*16.f, i*16.f + 1, m_grid_fg);
                vtx++;
                *vtx = s_grid_vertex(j*16.f, (i+1)*16.f, m_grid_fg);
                vtx++;
            }
        }
    }

    // create circle VB
    THR(m_pd3dDevice->CreateVertexBuffer(sizeof(s_circle_vertex)*NUM_CIRCLE_VERTICES,
        D3DUSAGE_WRITEONLY, s_circle_vertex::FVF, D3DPOOL_MANAGED, &m_circle));
    {
        rt::vertex_lock<s_circle_vertex> lock(m_circle);
        s_circle_vertex *vtx = lock.data();
        for (UINT i = 0; i < NUM_CIRCLE_VERTICES; i++)
        {
            vtx[i] = s_circle_vertex(i*D3DX_PI*2.f/(NUM_CIRCLE_VERTICES-1),
                m_circle_fg);
        }
    }

    // Set the world matrix
    D3DXMATRIX matIdentity;
    D3DXMatrixIdentity( &matIdentity );
    m_pd3dDevice->SetTransform( D3DTS_WORLD,  &matIdentity );

    // Set up our view matrix. A view matrix can be defined given an eye point,
    // a point to lookat, and a direction for which way is up. Here, we set the
    // eye five units back along the z-axis and up three units, look at the
    // origin, and define "up" to be in the y-direction.
    D3DXMATRIX matView;
    D3DXVECTOR3 vFromPt   = D3DXVECTOR3( 0.0f, 0.0f, -5.0f );
    D3DXVECTOR3 vLookatPt = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
    D3DXVECTOR3 vUpVec    = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
    D3DXMatrixLookAtLH( &matView, &vFromPt, &vLookatPt, &vUpVec );
    m_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

    // Set the projection matrix
    D3DXMATRIX matProj;
    FLOAT fAspect = ((FLOAT)m_d3dsdBackBuffer.Width) / m_d3dsdBackBuffer.Height;
    D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, fAspect, 1.0f, 100.0f );
    m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

    // Restore the font
    m_pFont->RestoreDeviceObjects();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
    // update grid
    // update circle
    if (m_show_circle)
    {
        // integrate();
    }

    // Update user input state
    UpdateInput( &m_UserInput );

    // Update the world state according to user input
    D3DXMATRIX matWorld;
    D3DXMATRIX matRotY;
    D3DXMATRIX matRotX;

    if( m_UserInput.bRotateLeft && !m_UserInput.bRotateRight )
        m_fWorldRotY += m_fElapsedTime;
    else if( m_UserInput.bRotateRight && !m_UserInput.bRotateLeft )
        m_fWorldRotY -= m_fElapsedTime;

    if( m_UserInput.bRotateUp && !m_UserInput.bRotateDown )
        m_fWorldRotX += m_fElapsedTime;
    else if( m_UserInput.bRotateDown && !m_UserInput.bRotateUp )
        m_fWorldRotX -= m_fElapsedTime;

    D3DXMatrixRotationX( &matRotX, m_fWorldRotX );
    D3DXMatrixRotationY( &matRotY, m_fWorldRotY );

    D3DXMatrixMultiply( &matWorld, &matRotX, &matRotY );
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: UpdateInput()
// Desc: Update the user input.  Called once per frame 
//-----------------------------------------------------------------------------
void CMyD3DApplication::UpdateInput( UserInput* pUserInput )
{
    pUserInput->bRotateUp    = ( m_bHasFocus && (GetAsyncKeyState( VK_UP )    & 0x8000) == 0x8000 );
    pUserInput->bRotateDown  = ( m_bHasFocus && (GetAsyncKeyState( VK_DOWN )  & 0x8000) == 0x8000 );
    pUserInput->bRotateLeft  = ( m_bHasFocus && (GetAsyncKeyState( VK_LEFT )  & 0x8000) == 0x8000 );
    pUserInput->bRotateRight = ( m_bHasFocus && (GetAsyncKeyState( VK_RIGHT ) & 0x8000) == 0x8000 );
}




//-----------------------------------------------------------------------------
// Name: RenderText()
// Desc: Renders stats and help text to the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RenderText()
{
    D3DCOLOR fontColor        = D3DCOLOR_ARGB(255,255,255,0);
    D3DCOLOR fontWarningColor = D3DCOLOR_ARGB(255,0,255,255);
    TCHAR szMsg[MAX_PATH] = TEXT("");

    // Output display stats
    FLOAT fNextLine = 40.0f; 

    lstrcpy( szMsg, m_strDeviceStats );
    fNextLine -= 20.0f;
    m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    lstrcpy( szMsg, m_strFrameStats );
    fNextLine -= 20.0f;
    m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    // Output statistics & help
    fNextLine = (FLOAT) m_d3dsdBackBuffer.Height; 

    wsprintf( szMsg, TEXT("Arrow keys: Up=%d Down=%d Left=%d Right=%d"), 
              m_UserInput.bRotateUp, m_UserInput.bRotateDown, m_UserInput.bRotateLeft, m_UserInput.bRotateRight );
    fNextLine -= 20.0f; m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    lstrcpy( szMsg, TEXT("Use arrow keys to rotate object") );
    fNextLine -= 20.0f; m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    lstrcpy( szMsg, TEXT("Press 'F2' to configure display") );
    fNextLine -= 20.0f; m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT msg, WPARAM wParam,
                                    LPARAM lParam )
{
    switch( msg )
    {
    case WM_PAINT:
        if( m_bLoadingApp )
        {
            // Draw on the window tell the user that the app is loading
            HDC hDC = GetDC( hWnd );
            TCHAR strMsg[MAX_PATH];
            wsprintf( strMsg, TEXT("Loading... Please wait") );
            RECT rct;
            GetClientRect( hWnd, &rct );
            DrawText( hDC, strMsg, -1, &rct, DT_CENTER|DT_VCENTER|DT_SINGLELINE );
            ReleaseDC( hWnd, hDC );
        }
        break;

    case WM_COMMAND:
        on_command(hWnd, wParam, lParam);
        break;
    }

    return CD3DApplication::MsgProc( hWnd, msg, wParam, lParam );
}




//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Invalidates device objects.  
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
    m_grid = 0;
    m_circle = 0;

    m_pFont->InvalidateDeviceObjects();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.  
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
    m_pFont->DeleteDeviceObjects();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
    // Cleanup D3D font
    SAFE_DELETE( m_pFont );

    // Write the settings to the registry
    WriteSettings();

    return S_OK;
}

////////////////////////////////////////////////////////////
// update_word
//
// Update an integer WORD value to and from an edit box,
// optionally displaying the value as hexadecimal.
//
void
update_word(HWND dialog, UINT control_id, WORD &val,
            bool hex, bool get_not_put)
{
    HWND edit = TWS(::GetDlgItem(dialog, control_id));
    if (get_not_put)
    {
        rt::tostringstream buff;
        if (hex)
        {
            buff << std::hex << std::showbase
                << std::setiosflags(std::ios::uppercase);
        }
        buff << val << std::ends;
        Edit_SetSel(edit, 0, -1);
        Edit_ReplaceSel(edit, buff.str().c_str());
    }
    else
    {
        std::vector<TCHAR> edit_text(20);
        {
            const WORD text_len = WORD(Edit_LineLength(edit, 0));
            edit_text.resize(text_len);

            if (text_len > 0)
            {
                ::memcpy(&edit_text[0], &text_len, sizeof(WORD));
                ::SendMessage(edit, EM_GETLINE, 0, LPARAM(&edit_text[0]));
                edit_text[text_len] = 0;
            }
        }

        rt::tistringstream stream(&edit_text[0]);
        DWORD value;
        stream >> std::resetiosflags(std::ios::basefield) >> value;
        if (!stream.fail())
        {
            val = value;
        }
    }
}

////////////////////////////////////////////////////////////
// CMyD3DApplication::update_line_pattern
//
// Line pattern dialog procedure; get/put line pattern on
// startup and shutdown of the dialog.
//
INT_PTR CALLBACK
CMyD3DApplication::update_line_pattern(HWND dialog, UINT msg,
                                       WPARAM wp, LPARAM lp)
{
    BOOL result = FALSE;

    // Static state for adapter/device/mode selection
    static D3DLINEPATTERN *line_pattern = NULL;

    // Handle the initialization message
    switch (msg)
    {
    case WM_INITDIALOG:
        line_pattern = reinterpret_cast<D3DLINEPATTERN *>(lp);

        // Set flag to update dialog controls below
        ::update_word(dialog, IDC_EDIT_REPEAT_COUNT,
            line_pattern->wRepeatFactor, false, true);
        ::update_word(dialog, IDC_EDIT_LINE_PATTERN,
            line_pattern->wLinePattern, true, true);
        result = TRUE;
        break;

    case WM_COMMAND:
        switch (LOWORD(wp))
        {
        case IDOK:
            ::update_word(dialog, IDC_EDIT_REPEAT_COUNT,
                line_pattern->wRepeatFactor, false, false);
            ::update_word(dialog, IDC_EDIT_LINE_PATTERN,
                line_pattern->wLinePattern, true, false);
            ::EndDialog(dialog, IDOK);
            result = TRUE;
            break;

        case IDCANCEL:
            ::EndDialog(dialog, IDCANCEL);
            result = TRUE;
            break;
        } // switch
        break;
    }

    return result;
}

////////////////////////////////////////////////////////////
// CMyD3DApplication::on_command
//
// WM_COMMAND handler -- update internal state based on GUI
// interaction.
//
void
CMyD3DApplication::on_command(HWND window, WPARAM wp, LPARAM lp)
{
    const HMENU menu = ::GetMenu(window);
    const UINT control = LOWORD(wp);

    switch (control)
    {
#define TOGGLE(id_, state_)                 \
    case id_:                               \
        state_ = !state_;                   \
        rt::check_menu(menu, id_, state_);  \
        break
    TOGGLE(IDM_OPTION_SHOW_TEXT, m_show_text);
    TOGGLE(IDM_OPTION_ANIMATE, m_animate);
    TOGGLE(IDM_OPTION_LAST_PIXEL, m_last_pixel);
    TOGGLE(IDM_OPTION_EDGE_ANTIALIAS, m_antialias_edge);
    TOGGLE(IDM_SCENE_GRID, m_show_grid);
    TOGGLE(IDM_SCENE_CIRCLE, m_show_circle);
#undef TOGGLE

    case IDM_OPTION_BACKGROUND:
        {
            rt::pauser block(*this);
            m_background = choose_color(window, m_background);
        }
        break;

    case IDM_OPTION_GRID_COLOR:
        {
            rt::pauser block(*this);
            m_grid_fg = choose_color(window, m_grid_fg);
            update_grid_color();
        }
        break;

    case IDM_OPTION_CIRCLE_COLOR:
        {
            rt::pauser block(*this);
            m_circle_fg = choose_color(window, m_circle_fg);
            update_circle_color();
        }
        break;

    case IDM_OPTION_LINE_PATTERN:
        {
            rt::pauser block(*this);
            DWORD line_pattern = m_line_pattern;
            if (IDOK ==
                DialogBoxParam(HINSTANCE(GetModuleHandle(NULL)),
                               MAKEINTRESOURCE(IDD_LINE_PATTERN), m_hWnd,
                               update_line_pattern,
                               reinterpret_cast<LPARAM>(&line_pattern)))
            {
                m_line_pattern = line_pattern;
            }
        }
        break;

    default:
        // all our control IDs are > 40006 and we should handle them all
        if (control > 40006)
        {
            ATLASSERT(false);
        }
    }
}
