////////////////////////////////////////////////////////////
// rt_Rasterize.cpp
//
// Demonstrate RS Edge Antialias, RS Last Pixel and
// RS Line Pattern rasterization state.
//
// Rich Thomson
//
#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <D3DX8.h>

#include "D3DApp.h"
#include "D3DFont.h"

#include "resource.h"
#include "rt_Rasterize.h"
#include "vertices.h"

#include "rt/hr.h"
#include "rt/misc.h"
#include "rt/states.h"
#include "rt/vertexbuf.h"

////////////////////////////////////////////////////////////
// s_circle_vertex::FVF
// s_grid_vertex::FVF
//
// FVF codes for the circle and grid vertices.
//
const DWORD
s_circle_vertex::FVF = D3DFVF_XYZ | D3DFVF_DIFFUSE;
const DWORD
s_grid_vertex::FVF = D3DFVF_XYZRHW | D3DFVF_DIFFUSE;


////////////////////////////////////////////////////////////
// Render()
//
// Draw a scene consisting of a screen-space grid every 16
// pixels and a world space circle, initially in the xy
// plane.
//
HRESULT
CMyD3DApplication::Render()
{
    // Clear the viewport: no Z buffer is used
    THR(m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET,
        m_background, 1.0f, 0L));

    THR(m_pd3dDevice->BeginScene());

    // set rasterization state for lines:
    // RS Edge Antialias based on GUI state
    // RS Last Pixel based on GUI state
    // RS Line Pattern to default of 0 for grid
    // alpha blending,
    // no lighting or Z buffering
    rt::s_rs states[] =
    {
        D3DRS_EDGEANTIALIAS,
            m_can_antialias_edge && m_antialias_edge,
        D3DRS_LASTPIXEL, m_last_pixel,
        D3DRS_LINEPATTERN, 0,
        D3DRS_ALPHABLENDENABLE, TRUE,
        D3DRS_SRCBLEND, D3DBLEND_SRCALPHA,
        D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA,
        D3DRS_ZENABLE, D3DZB_FALSE,
        D3DRS_LIGHTING, FALSE
    };
    rt::set_states(m_pd3dDevice, states, NUM_OF(states));

    // draw the screen space grid if requested
    if (m_show_grid)
    {
        THR(m_pd3dDevice->
            SetVertexShader(s_grid_vertex::FVF));
        THR(m_pd3dDevice->SetStreamSource(0, m_grid,
            sizeof(s_grid_vertex)));
        THR(m_pd3dDevice->DrawPrimitive(D3DPT_LINELIST, 0,
            m_num_grid_segments));
    }

    // draw world space circle if requested
    if (m_show_circle)
    {
        // set the line pattern for the circle
        THR(m_pd3dDevice->SetRenderState(D3DRS_LINEPATTERN,
            m_line_pattern));
        THR(m_pd3dDevice->
            SetVertexShader(s_circle_vertex::FVF));
        THR(m_pd3dDevice->SetStreamSource(0, m_circle,
            sizeof(s_circle_vertex)));
        THR(m_pd3dDevice->DrawPrimitive(D3DPT_LINESTRIP, 0,
            NUM_CIRCLE_VERTICES-1));
    }

    // Render stats and help text
    if (m_show_text)
    {
        RenderText();
    }

    THR(m_pd3dDevice->EndScene());

    return S_OK;
}

////////////////////////////////////////////////////////////
// update_grid_color
//
// Store the new grid color in the grid vertices.
//
void
CMyD3DApplication::update_grid_color()
{
    rt::vertex_lock<s_grid_vertex> lock(m_grid);
    s_grid_vertex *vtx = lock.data();
    for (UINT i = 0; i < m_num_grid_segments*2; i++)
    {
        vtx[i].m_fg = m_grid_fg;
    }
}

////////////////////////////////////////////////////////////
// update_circle_color
//
// Store the new circle color in the circle vertices.
//
void
CMyD3DApplication::update_circle_color()
{
    rt::vertex_lock<s_circle_vertex> lock(m_circle);
    s_circle_vertex *vtx = lock.data();
    for (UINT i = 0; i < NUM_CIRCLE_VERTICES; i++)
    {
        vtx[i].m_fg = m_circle_fg;
    }
}
